import numpy as np
import matplotlib.pyplot as plt
from matplotlib.colors import LogNorm, Normalize
from .load import load


class ImageViewer():

    def __init__(self, x, y, z, filename=None, colormap="viridis", vmin=None,
                 vmax=None, log=False, side_panels=True, clab="", imstart=0.1,
                 ax=None):

        cbsize = 0.02
        figend = 0.95
        imsize = figend - 2.0 * imstart
        ytitle = 1.08
        if side_panels:
            plsize = 0.25 * imsize
            imsize = imsize - plsize
            ytitle = 1.38

        if vmin is None:
            vmin = np.nanmin(np.clip(z, 1.0, None) if log else z)
        if vmax is None:
            vmax = np.nanmax(np.clip(z, 1.0, None) if log else z)

        norm = LogNorm(vmin=vmin, vmax=vmax) if log else Normalize(vmin=vmin, vmax=vmax)

        self.side_panels = side_panels
        if ax is None:
            self.fig = plt.figure(figsize=(8, 8))
            self.ax1 = self.fig.add_axes([imstart, imstart, imsize, imsize])
            if self.side_panels:
                self.ax2 = self.fig.add_axes([1.0 - imstart, imstart, cbsize, imsize])
            else:
                self.ax2 = self.fig.add_axes([imstart + imsize + cbsize, imstart, cbsize, imsize])
        else:
            self.ax1 = ax
            self.ax2 = None

        self.im = self.ax1.imshow(z, origin="lower", aspect="auto", interpolation="none",
                                  norm=norm, extent=[x[0], x[-1], y[0], y[-1]])

        self.cb = plt.colorbar(self.im, ax=self.ax1, cax=self.ax2)
        if self.side_panels:
            self.cb.ax.set_xlabel(clab)
            self.cb.ax.xaxis.set_label_position("top")
        else:
            self.cb.ax.set_ylabel(clab)
        self.ax1.set_xlabel("x position [m]")
        self.ax1.set_ylabel("y position [m]")

        if self.side_panels:
            z_sumx = np.sum(z, axis=1)
            z_sumy = np.sum(z, axis=0)
            self.ax3 = self.fig.add_axes([imstart, imstart + imsize, imsize, plsize], sharex=self.ax1)
            self.ax4 = self.fig.add_axes([imstart + imsize, imstart, plsize, imsize], sharey=self.ax1)
            self.ax3.grid(True, color="lightgray", linestyle="dotted")
            self.ax4.grid(True, color="lightgray", linestyle="dotted")
            self.ax3.set_axisbelow(True)
            self.ax4.set_axisbelow(True)
            self.plot_x, = self.ax3.plot(x, z_sumy)
            self.plot_y, = self.ax4.plot(z_sumx, y)
            self.ax3.set_xlim([x[0], x[-1]])
            self.ax4.set_ylim([y[0], y[-1]])
            self.ax3.xaxis.tick_top()
            self.ax3.xaxis.set_label_position("top")
            self.ax4.yaxis.tick_right()
            self.ax3.set_xlabel("x position [m]")
            self.ax3.set_ylabel("Counts")
            self.ax4.set_xlabel("Counts")
            for tick in self.ax4.get_xticklabels():
                tick.set_rotation(-90)
            for tick in self.ax4.get_yticklabels():
                tick.set_rotation(-90)
            if log:
                self.ax3.set_yscale("log", nonposy="clip")
                self.ax4.set_xscale("log", nonposx="clip")
            self.ax3.set_title(filename.split("/")[-1], y=ytitle,
                               bbox=dict(facecolor="none", edgecolor="grey",
                                         boxstyle="round"))
        elif filename is not None:
            self.ax1.set_title(filename.split("/")[-1], y=ytitle,
                               bbox=dict(facecolor="none", edgecolor="grey",
                                         boxstyle="round"))
        return


def image(filename=None, data=None, colormap="viridis", vmin=None, vmax=None,
          log=False, side_panels=True, save=None, ax=None, **kwargs):
    """
    Make a 2D image of the detector counts
    """

    show = ax is None

    if data is None:
        data = load(filename, ids=True, **kwargs)

    z, edges = np.histogram(data.ids,
                            bins=np.arange(-0.5, data.nx * data.ny + 0.5))
    z = z.reshape(data.ny, data.nx)

    imv = ImageViewer(data.x[0, :], data.y[:, 0], z, filename=filename,
                      colormap=colormap, vmin=vmin, vmax=vmax, log=log,
                      side_panels=side_panels, clab="Counts", ax=ax)

    if show:
        if save is not None:
            imv.fig.savefig(save, bbox_inches="tight")
        else:
            imv.fig.show()



from .tof import tof
from .image import image
from .xy_vs_tof import x_vs_tof, y_vs_tof
from .slicer import slicer
from .volume import volume
from .instrument_view import instrument_view
from .load import load
from .preview import preview



import numpy as np
import matplotlib.pyplot as plt
from .load import load
from .image import image
from .tof import tof
from .xy_vs_tof import x_vs_tof, y_vs_tof


def preview(filename=None, data=None, save=None, log=False):
    """
    Make a 4-panel file preview
    """

    if data is None:
        data = load(filename, ids=True, tofs=True, convert_ids=False)

    fig, ax = plt.subplots(2, 2, figsize=(12, 9))
    fig.suptitle(filename.split("/")[-1])

    tof(data=data, ax=ax[0][0])
    x_vs_tof(data=data, ax=ax[0][1], log=log)
    image(data=data, ax=ax[1][0], side_panels=False, log=log)
    y_vs_tof(data=data, ax=ax[1][1], log=log)

    if save is not None:
        fig.savefig(save, bbox_inches="tight")
    else:
        plt.show()



import numpy as np
import matplotlib.pyplot as plt
from .load import load


def xy_vs_tof(filename=None, data=None, colormap="viridis", vmin=None,
              vmax=None, log=False, nbins=512, save=None, axis=0,
              ax=None, **kwargs):
    """
    Make a x or y vs tof image
    """

    if data is None:
        data = load(filename, ids=True, tofs=True, **kwargs)

    t = np.linspace(0.0, 7.2e4, nbins + 1)
    z, xe, ye = np.histogram2d(data.ids, data.tofs/1.0e3,
                               bins=[np.arange(-0.5, data.nx * data.ny + 0.5),
                                     t])
    z = z.reshape(data.ny, data.nx, nbins)
    z = np.sum(z, axis=axis)
    clab = "Counts"
    if log:
        with np.errstate(divide="ignore", invalid="ignore"):
            z = np.log10(z)
        clab = "log({})".format(clab)

    show = True
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        show = False

    im = ax.imshow(z, origin="lower", aspect="auto", interpolation="none",
                   vmin=vmin, vmax=vmax,
                   extent=[t[0], t[-1], data.x[0, 0], data.x[0, -1]])
    cb = plt.colorbar(im, ax=ax)
    cb.ax.set_ylabel(clab)
    ax.set_xlabel("Time-of-flight [microseconds]")
    ax.set_ylabel("x position [m]")
    if filename is not None:
        ax.set_title(filename.split("/")[-1])

    if show:
        if save is not None:
            fig.savefig(save, bbox_inches="tight")
        else:
            plt.show()


def x_vs_tof(filename=None, data=None, colormap="viridis", vmin=None,
             vmax=None, log=False, nbins=512, save=None, transpose=False,
             ax=None, **kwargs):
    """
    Make a x vs tof image
    """

    xy_vs_tof(filename=filename, data=data, colormap=colormap, vmin=vmin,
              vmax=vmax, log=log, nbins=nbins, save=save, axis=int(transpose),
              ax=ax, **kwargs)


def y_vs_tof(filename=None, data=None, colormap="viridis", vmin=None,
             vmax=None, log=False, nbins=512, save=None, transpose=False,
             ax=None, **kwargs):
    """
    Make a y vs tof image
    """

    xy_vs_tof(filename=filename, data=data, colormap=colormap, vmin=vmin,
              vmax=vmax, log=log, nbins=nbins, save=save,
              axis=int(not transpose), ax=ax, **kwargs)



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider
from .load import load
from .image import ImageViewer

class Slicer(object):

    def __init__(self, x, y, z, tof, dt=0.0, filename="", colormap="viridis",
                 vmin=None, vmax=None, log=False, side_panels=True, clab=""):

        # Initial index
        self.ind = 0

        # Use ImageViewer class
        self.imv = ImageViewer(x, y, z[:, :, self.ind], filename=filename,
                      colormap=colormap, vmin=vmin, vmax=vmax, log=log,
                      side_panels=side_panels, clab=clab, imstart=0.12)

        self.x = x
        self.y = y
        self.z = z
        self.tof = tof
        self.dt = dt
        self.nslices = z.shape[-1]

        # Set limits
        if self.imv.side_panels:
            self.z_sumy = np.sum(z, axis=0)
            self.z_sumx = np.sum(z, axis=1)
            self.imv.ax3.set_ylim(top=np.nanmax(self.z_sumy))
            self.imv.ax4.set_xlim(right=np.nanmax(self.z_sumx))

        # Connect canvas to scroll
        self.imv.fig.canvas.mpl_connect("scroll_event", self.onscroll)

        # Add mpl slider widget
        self.ax_slider = self.imv.fig.add_axes([0.23, 0.02, 0.56, 0.03])
        self.slider = Slider(self.ax_slider, "Tof [us]", tof.min(), tof.max(),
                             valinit=tof.min())
        # Connect slider to function
        self.slider.on_changed(self.update)

    def onscroll(self, event):
        """
        Allow moving the slider with the mouse wheel
        """
        if event.button == "up":
            self.ind = np.clip(self.ind + 1, 0, self.nslices - 1)
        else:
            self.ind = np.clip(self.ind - 1, 0, self.nslices - 1)
        self.slider.set_val(self.ind * self.dt)

    def update(self, val=None):
        """
        Update the image with new data according to slider value
        """
        self.ind = np.clip(int(round(val/self.dt)), 0, self.nslices - 1)
        self.imv.im.set_data(self.z[:, :, self.ind])
        if self.imv.side_panels:
            self.imv.plot_x.set_data(self.x, self.z_sumy[:, self.ind])
            self.imv.plot_y.set_data(self.z_sumx[:, self.ind], self.y)


def slicer(filename=None, data=None, colormap="viridis", vmin=None, vmax=None,
           log=False, nbins=256, transpose=False, side_panels=True, **kwargs):
    """
    Make a 2D image viewer with a slider to navigate in the Tof dimension.
    You can also scroll with the mouse wheel to change the slider position.
    """

    if data is None:
        data = load(filename, ids=True, tofs=True, **kwargs)

    t = np.linspace(0.0, 7.2e4, nbins + 1)
    z, xe, ye = np.histogram2d(data.ids, data.tofs/1.0e3,
                               bins=[np.arange(-0.5, data.nx * data.ny + 0.5),
                                     t])
    z = z.reshape(data.ny, data.nx, nbins)
    # Transpose should be True for old December 2018 files
    if transpose:
        z = np.transpose(z, axes=[1, 0, 2])

    # Create slicer object
    sl = Slicer(data.x[0, :], data.y[:, 0], z, 0.5*(t[1:] + t[:-1]),
                dt=t[1]-t[0], filename=filename, colormap=colormap, vmin=vmin,
                vmax=vmax, log=log, side_panels=side_panels, clab="Counts")
    # Assume we are always opening in interactive mode
    sl.imv.fig.show()

    return sl



import numpy as np
import matplotlib.pyplot as plt
from .load import load


def tof(filename=None, data=None, xmin=None, xmax=None, logx=False, logy=False,
        logxy=False, nbins=512, save=None, ax=None, **kwargs):
    """
    Make counts vs tof histogram
    """

    if data is None:
        data = load(filename, tofs=True, **kwargs)

    tofs = data.tofs / 1.0e3
    if xmin is None:
        xmin = np.amin(tofs)
    if xmax is None:
        xmax = np.amax(tofs)

    show = True
    if ax is None:
        fig = plt.figure()
        ax = fig.add_subplot(111)
    else:
        show = False

    if logx or logxy:
        ax.set_xscale("log", nonposx='clip')
    if logy or logxy:
        ax.set_yscale("log", nonposy='clip')
    ax.hist(tofs, bins=np.linspace(xmin, xmax, nbins + 1))
    ax.set_xlabel("Time-of-flight [microseconds]")
    ax.set_ylabel("Counts")
    if filename is not None:
        ax.set_title(filename.split("/")[-1])

    if show:
        if save is not None:
            fig.savefig(save, bbox_inches="tight")
        else:
            plt.show()



import numpy as np
from .load import load


def volume(filename=None, data=None, colormap="viridis", nbins=256, save=None,
           **kwargs):
    """
    Make a 3D volume rendering using ipyvolume
    """

    try:
        import ipyvolume as ipv
    except ImportError:
        print("Volume rendering makes use of the ipyvolume package which was "
              "not found on this system. Install using pip install ipyvolume.")

    if data is None:
        data = load(filename, ids=True, tofs=True, **kwargs)

    t = np.linspace(0.0, 7.2e4, nbins + 1)
    z, xe, ye = np.histogram2d(data.ids, data.tofs/1.0e3,
                               bins=[np.arange(-0.5, data.nx * data.ny + 0.5),
                                     t])
    z = np.transpose(z.reshape(data.ny, data.nx, nbins), axes=[2, 1, 0])
    ipv.quickvolshow(z, extent=[[data.x[0, 0]*100.0, data.x[0, -1]*100.0],
                                [data.y[0, 0]*100.0, data.y[-1, 0]*100.0],
                                [t[0], t[-1]]])
    ipv.pylab.xlabel("x [cm]")
    ipv.pylab.ylabel("y [cm]")
    ipv.pylab.zlabel("Tof [us]")
    if save is not None:
        ipv.pylab.save(save)
    else:
        ipv.show()

    return



import h5py
import numpy as np


class NxsData:
    """
    Class to hold data in Nexus file
    """

    def __init__(self, fields=None):

        for key in fields.keys():
            setattr(self, key, fields[key]["data"])

        return


def load(filename, ids=False, tofs=False, entry="/", verbose=False,
         convert_ids=True):
    """
    Load a hdf/nxs file and return required information.
    Note that the patterns are listed in order of preference,
    i.e. if more than one is present in the file, the data will be read
    from the first one found.
    """

    fields = {"x": {"pattern": ["/x_pixel_offset"],
                    "entry": None,
                    "data": None,
                    "dtype": np.float64},
              "y": {"pattern": ["/y_pixel_offset"],
                    "entry": None,
                    "data": None,
                    "dtype": np.float64}}

    if ids:
        fields["ids"] = {"pattern": ["/event_data/event_id",
                                     "/raw_event_data/event_id"],
                         "entry": None,
                         "data": None,
                         "dtype": np.int32}
    if tofs:
        fields["tofs"] = {"pattern": ["/event_data/event_time_offset",
                                      "/raw_event_data/event_time_offset"],
                          "entry": None,
                          "data": None,
                          "dtype": np.float64}

    with h5py.File(filename, "r", libver='latest', swmr=True) as f:

        contents = []
        f[entry].visit(contents.append)

        for item in contents:
            for key in fields.keys():
                if fields[key]["entry"] is None:
                    for p in fields[key]["pattern"]:
                        if item.endswith(p):
                            fields[key]["entry"] = item

        for key in fields.keys():
            fields[key]["data"] = np.array(f[fields[key]["entry"]][...],
                                           dtype=fields[key]["dtype"],
                                           copy=True)

    # Store the detector size in pixels
    nx, ny = np.shape(fields["x"]["data"])
    fields["nx"] = {"entry": None, "data": nx}
    fields["ny"] = {"entry": None, "data": ny}

    if ids and convert_ids and (np.amax(fields["ids"]["data"]) >= nx*ny):
        print("Warning: maximum id exceeds total number of pixels. "
              "Attempting to fix ids...")
        fields["ids"]["data"] = __convert_id(fields["ids"]["data"], nx)

    if verbose:
        for key in fields.keys():
            print("Loaded {} from: {}".format(key, fields[key]["entry"]))
            print("  - Data size: {} : Min={} , "
                  "Max={}".format(np.shape(fields[key]["data"]),
                                  np.amin(fields[key]["data"]),
                                  np.amax(fields[key]["data"])))

    return NxsData(fields)


def __convert_id(ids, nx, id_offset=0):
    """
    Attempt to convert the event ids if their range exceeds the maximum index
    allowed by the pixel geometry.
    """
    x = np.bitwise_and(ids, 0xFFFF)
    y = np.right_shift(ids, 16)
    return id_offset + x + (nx * y)
